from __future__ import annotations

import json
import random
import time
import uuid
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Mapping, MutableMapping, Optional
from urllib.parse import quote, urlencode
from urllib import error, request

if TYPE_CHECKING:
    from .types import (
        ControlAlertDeliveriesQuery,
        ControlAlertRouteInput,
        ControlAlertRoutesQuery,
        ControlAlertRouteStatusInput,
        ControlApiKeyInput,
        ControlApiKeysQuery,
        ControlApiKeysStaleQuery,
        ControlApiKeyRotateInput,
        ControlAuditEventsQuery,
        ControlIncidentPublishJobInput,
        ControlIncidentPublishJobsQuery,
        ControlIncidentPublishReplayInput,
        ControlIncidentPublishRollupQuery,
        ControlIncidentPublishSloQuery,
        ControlProjectInput,
        ControlTenantDiagnosticsQuery,
        ControlTenantInput,
        ControlTenantKeyUsageQuery,
        ControlTenantQuotaInput,
        ControlTenantsQuery,
        ControlTenantTimeseriesQuery,
        BackendCapabilityErrorDetails,
        MemoryEventWriteInput,
        MemoryFindInput,
        MemoryPackExportInput,
        MemoryPackImportInput,
        MemoryRecallInput,
        MemoryRecallTextInput,
        MemorySessionCreateInput,
        MemorySessionEventsListInput,
        MemoryWriteInput,
        RulesEvaluateInput,
        ToolsDecisionInput,
        ToolsFeedbackInput,
        ToolsSelectInput,
        ShadowDualWriteStrictFailureDetails,
    )


@dataclass
class RetryPolicy:
    max_retries: int = 2
    base_delay_s: float = 0.2
    max_delay_s: float = 2.0
    jitter_ratio: float = 0.2


class AionisApiError(Exception):
    def __init__(
        self,
        status: int,
        code: str,
        message: str,
        details: Any = None,
        request_id: Optional[str] = None,
        issues: Optional[list[dict[str, Any]]] = None,
    ) -> None:
        super().__init__(message)
        self.status = status
        self.code = code
        self.details = details
        self.request_id = request_id
        self.issues = issues


class AionisNetworkError(Exception):
    def __init__(self, message: str, request_id: Optional[str] = None) -> None:
        super().__init__(message)
        self.request_id = request_id


def parse_backend_capability_error_details(details: Any) -> Optional["BackendCapabilityErrorDetails"]:
    if not isinstance(details, dict):
        return None
    capability = details.get("capability")
    if not isinstance(capability, str) or not capability.strip():
        return None
    out: Dict[str, Any] = {"capability": capability}
    backend = details.get("backend")
    if isinstance(backend, str):
        out["backend"] = backend
    failure_mode = details.get("failure_mode")
    if failure_mode in ("hard_fail", "soft_degrade"):
        out["failure_mode"] = failure_mode
    degraded_mode = details.get("degraded_mode")
    if isinstance(degraded_mode, str):
        out["degraded_mode"] = degraded_mode
    fallback_applied = details.get("fallback_applied")
    if isinstance(fallback_applied, bool):
        out["fallback_applied"] = fallback_applied
    return out  # type: ignore[return-value]


def is_backend_capability_unsupported_error(err: BaseException) -> bool:
    if not isinstance(err, AionisApiError):
        return False
    if err.code != "backend_capability_unsupported":
        return False
    parsed = parse_backend_capability_error_details(err.details)
    if parsed is None:
        return False
    err.details = parsed
    return True


def parse_shadow_dual_write_strict_failure_details(details: Any) -> Optional["ShadowDualWriteStrictFailureDetails"]:
    if not isinstance(details, dict):
        return None
    if details.get("capability") != "shadow_mirror_v2":
        return None
    out: Dict[str, Any] = {"capability": "shadow_mirror_v2"}
    failure_mode = details.get("failure_mode")
    if failure_mode in ("hard_fail", "soft_degrade"):
        out["failure_mode"] = failure_mode
    degraded_mode = details.get("degraded_mode")
    if degraded_mode in ("capability_unsupported", "mirror_failed"):
        out["degraded_mode"] = degraded_mode
    fallback_applied = details.get("fallback_applied")
    if isinstance(fallback_applied, bool):
        out["fallback_applied"] = fallback_applied
    strict = details.get("strict")
    if isinstance(strict, bool):
        out["strict"] = strict
    mirrored = details.get("mirrored")
    if isinstance(mirrored, bool):
        out["mirrored"] = mirrored
    error_msg = details.get("error")
    if isinstance(error_msg, str):
        out["error"] = error_msg
    return out  # type: ignore[return-value]


def is_shadow_dual_write_strict_failure_error(err: BaseException) -> bool:
    if not isinstance(err, AionisApiError):
        return False
    if err.code != "shadow_dual_write_strict_failure":
        return False
    parsed = parse_shadow_dual_write_strict_failure_details(err.details)
    if parsed is None:
        return False
    err.details = parsed
    return True


def _join_url(base_url: str, path: str) -> str:
    return f"{base_url.rstrip('/')}/{path.lstrip('/')}"


def _should_retry_status(status: int) -> bool:
    return status == 429 or status >= 500


def _parse_retry_after_seconds(headers: Mapping[str, str]) -> Optional[float]:
    raw = headers.get("retry-after")
    if raw is None:
        return None
    try:
        sec = float(raw)
        return sec if sec > 0 else None
    except ValueError:
        return None


def _compute_backoff_seconds(policy: RetryPolicy, attempt: int) -> float:
    exp = policy.base_delay_s * (2 ** max(0, attempt - 1))
    capped = min(exp, policy.max_delay_s)
    spread = capped * policy.jitter_ratio
    if spread <= 0:
        return max(0.001, capped)
    jitter = random.uniform(-spread, spread)
    return max(0.001, capped + jitter)


def _parse_body(raw: bytes, content_type: str) -> Any:
    if not raw:
        return None
    if "application/json" in (content_type or "").lower():
        try:
            return json.loads(raw.decode("utf-8"))
        except json.JSONDecodeError:
            return raw.decode("utf-8", errors="replace")
    return raw.decode("utf-8", errors="replace")


def _has_header(headers: Mapping[str, str], name: str) -> bool:
    want = name.lower()
    return any(str(k).lower() == want for k in headers.keys())


def _as_retry_policy(value: Optional[Mapping[str, Any]], base: RetryPolicy) -> RetryPolicy:
    if value is None:
        value = {}
    max_retries = int(value.get("max_retries", base.max_retries))
    base_delay_s = float(value.get("base_delay_s", base.base_delay_s))
    max_delay_s = float(value.get("max_delay_s", base.max_delay_s))
    jitter_ratio = float(value.get("jitter_ratio", base.jitter_ratio))

    return RetryPolicy(
        max_retries=max(0, min(10, max_retries)),
        base_delay_s=max(0.001, min(30.0, base_delay_s)),
        max_delay_s=max(0.001, min(60.0, max_delay_s)),
        jitter_ratio=max(0.0, min(1.0, jitter_ratio)),
    )


class AionisClient:
    def __init__(
        self,
        *,
        base_url: str,
        timeout_s: float = 10.0,
        retry: Optional[Mapping[str, Any]] = None,
        default_headers: Optional[Mapping[str, str]] = None,
        admin_token: Optional[str] = None,
        api_key: Optional[str] = None,
        auth_bearer: Optional[str] = None,
    ) -> None:
        self.base_url = base_url.rstrip("/")
        self.timeout_s = max(0.001, timeout_s)
        self.retry = _as_retry_policy(retry, RetryPolicy())
        self.default_headers = dict(default_headers or {})
        self.admin_token = admin_token
        self.api_key = api_key
        self.auth_bearer = auth_bearer

    def write(self, payload: "MemoryWriteInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/write", payload, request_options)

    def recall(self, payload: "MemoryRecallInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/recall", payload, request_options)

    def recall_text(self, payload: "MemoryRecallTextInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/recall_text", payload, request_options)

    def find(self, payload: "MemoryFindInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/find", payload, request_options)

    def create_session(self, payload: "MemorySessionCreateInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/sessions", payload, request_options)

    def write_event(self, payload: "MemoryEventWriteInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/events", payload, request_options)

    def list_session_events(
        self,
        session_id: str,
        query: Optional["MemorySessionEventsListInput"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        sid = str(session_id or "").strip()
        if not sid:
            raise ValueError("session_id is required")
        path = f"/v1/memory/sessions/{quote(sid, safe='')}/events"
        return self._request(path, query or {}, request_options, method="GET")

    def pack_export(self, payload: "MemoryPackExportInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/packs/export", payload, request_options)

    def pack_import(self, payload: "MemoryPackImportInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/packs/import", payload, request_options)

    def rules_evaluate(self, payload: "RulesEvaluateInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/rules/evaluate", payload, request_options)

    def tools_select(self, payload: "ToolsSelectInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/tools/select", payload, request_options)

    def tools_decision(self, payload: "ToolsDecisionInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/tools/decision", payload, request_options)

    def tools_feedback(self, payload: "ToolsFeedbackInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/memory/tools/feedback", payload, request_options)

    def health(self, **request_options: Any) -> Dict[str, Any]:
        return self._request("/health", {}, request_options, method="GET")

    def get_capability_contract(self, **request_options: Any) -> Dict[str, Any]:
        out = self.health(**request_options)
        data = out.get("data") if isinstance(out, dict) else None
        contract: Dict[str, Any] = {}
        if isinstance(data, dict):
            maybe = data.get("memory_store_capability_contract")
            if isinstance(maybe, dict):
                contract = maybe
        return {
            "data": contract,
            "status": out.get("status"),
            "request_id": out.get("request_id"),
        }

    def control_upsert_tenant(self, payload: "ControlTenantInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/admin/control/tenants", payload, request_options)

    def control_list_tenants(
        self,
        query: Optional["ControlTenantsQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        return self._request("/v1/admin/control/tenants", query or {}, request_options, method="GET")

    def control_upsert_project(self, payload: "ControlProjectInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/admin/control/projects", payload, request_options)

    def control_create_api_key(self, payload: "ControlApiKeyInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/admin/control/api-keys", payload, request_options)

    def control_list_api_keys(
        self,
        query: Optional["ControlApiKeysQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        return self._request("/v1/admin/control/api-keys", query or {}, request_options, method="GET")

    def control_list_stale_api_keys(
        self,
        query: Optional["ControlApiKeysStaleQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        return self._request("/v1/admin/control/api-keys/stale", query or {}, request_options, method="GET")

    def control_revoke_api_key(self, key_id: str, **request_options: Any) -> Dict[str, Any]:
        sid = str(key_id or "").strip()
        if not sid:
            raise ValueError("key_id is required")
        path = f"/v1/admin/control/api-keys/{quote(sid, safe='')}/revoke"
        return self._request(path, {}, request_options)

    def control_rotate_api_key(
        self,
        key_id: str,
        payload: Optional["ControlApiKeyRotateInput"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        sid = str(key_id or "").strip()
        if not sid:
            raise ValueError("key_id is required")
        path = f"/v1/admin/control/api-keys/{quote(sid, safe='')}/rotate"
        return self._request(path, payload or {}, request_options)

    def control_create_alert_route(self, payload: "ControlAlertRouteInput", **request_options: Any) -> Dict[str, Any]:
        return self._request("/v1/admin/control/alerts/routes", payload, request_options)

    def control_list_alert_routes(
        self,
        query: Optional["ControlAlertRoutesQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        return self._request("/v1/admin/control/alerts/routes", query or {}, request_options, method="GET")

    def control_update_alert_route_status(
        self,
        route_id: str,
        payload: "ControlAlertRouteStatusInput",
        **request_options: Any,
    ) -> Dict[str, Any]:
        sid = str(route_id or "").strip()
        if not sid:
            raise ValueError("route_id is required")
        path = f"/v1/admin/control/alerts/routes/{quote(sid, safe='')}/status"
        return self._request(path, payload, request_options)

    def control_list_alert_deliveries(
        self,
        query: Optional["ControlAlertDeliveriesQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        return self._request("/v1/admin/control/alerts/deliveries", query or {}, request_options, method="GET")

    def control_enqueue_incident_publish_job(
        self,
        payload: "ControlIncidentPublishJobInput",
        **request_options: Any,
    ) -> Dict[str, Any]:
        return self._request("/v1/admin/control/incident-publish/jobs", payload, request_options)

    def control_list_incident_publish_jobs(
        self,
        query: Optional["ControlIncidentPublishJobsQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        return self._request("/v1/admin/control/incident-publish/jobs", query or {}, request_options, method="GET")

    def control_replay_incident_publish_jobs(
        self,
        payload: "ControlIncidentPublishReplayInput",
        **request_options: Any,
    ) -> Dict[str, Any]:
        return self._request("/v1/admin/control/incident-publish/jobs/replay", payload, request_options)

    def control_upsert_tenant_quota(
        self,
        tenant_id: str,
        payload: "ControlTenantQuotaInput",
        **request_options: Any,
    ) -> Dict[str, Any]:
        sid = str(tenant_id or "").strip()
        if not sid:
            raise ValueError("tenant_id is required")
        path = f"/v1/admin/control/tenant-quotas/{quote(sid, safe='')}"
        return self._request(path, payload, request_options, method="PUT")

    def control_get_tenant_quota(self, tenant_id: str, **request_options: Any) -> Dict[str, Any]:
        sid = str(tenant_id or "").strip()
        if not sid:
            raise ValueError("tenant_id is required")
        path = f"/v1/admin/control/tenant-quotas/{quote(sid, safe='')}"
        return self._request(path, {}, request_options, method="GET")

    def control_delete_tenant_quota(self, tenant_id: str, **request_options: Any) -> Dict[str, Any]:
        sid = str(tenant_id or "").strip()
        if not sid:
            raise ValueError("tenant_id is required")
        path = f"/v1/admin/control/tenant-quotas/{quote(sid, safe='')}"
        return self._request(path, {}, request_options, method="DELETE")

    def control_list_audit_events(
        self,
        query: Optional["ControlAuditEventsQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        return self._request("/v1/admin/control/audit-events", query or {}, request_options, method="GET")

    def control_get_tenant_dashboard(self, tenant_id: str, **request_options: Any) -> Dict[str, Any]:
        sid = str(tenant_id or "").strip()
        if not sid:
            raise ValueError("tenant_id is required")
        path = f"/v1/admin/control/dashboard/tenant/{quote(sid, safe='')}"
        return self._request(path, {}, request_options, method="GET")

    def control_get_tenant_diagnostics(
        self,
        tenant_id: str,
        query: Optional["ControlTenantDiagnosticsQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        sid = str(tenant_id or "").strip()
        if not sid:
            raise ValueError("tenant_id is required")
        path = f"/v1/admin/control/diagnostics/tenant/{quote(sid, safe='')}"
        return self._request(path, query or {}, request_options, method="GET")

    def control_get_tenant_incident_publish_rollup(
        self,
        tenant_id: str,
        query: Optional["ControlIncidentPublishRollupQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        sid = str(tenant_id or "").strip()
        if not sid:
            raise ValueError("tenant_id is required")
        path = f"/v1/admin/control/dashboard/tenant/{quote(sid, safe='')}/incident-publish-rollup"
        return self._request(path, query or {}, request_options, method="GET")

    def control_get_tenant_incident_publish_slo(
        self,
        tenant_id: str,
        query: Optional["ControlIncidentPublishSloQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        sid = str(tenant_id or "").strip()
        if not sid:
            raise ValueError("tenant_id is required")
        path = f"/v1/admin/control/dashboard/tenant/{quote(sid, safe='')}/incident-publish-slo"
        return self._request(path, query or {}, request_options, method="GET")

    def control_get_tenant_timeseries(
        self,
        tenant_id: str,
        query: Optional["ControlTenantTimeseriesQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        sid = str(tenant_id or "").strip()
        if not sid:
            raise ValueError("tenant_id is required")
        path = f"/v1/admin/control/dashboard/tenant/{quote(sid, safe='')}/timeseries"
        return self._request(path, query or {}, request_options, method="GET")

    def control_get_tenant_key_usage(
        self,
        tenant_id: str,
        query: Optional["ControlTenantKeyUsageQuery"] = None,
        **request_options: Any,
    ) -> Dict[str, Any]:
        sid = str(tenant_id or "").strip()
        if not sid:
            raise ValueError("tenant_id is required")
        path = f"/v1/admin/control/dashboard/tenant/{quote(sid, safe='')}/key-usage"
        return self._request(path, query or {}, request_options, method="GET")

    def _request(
        self,
        path: str,
        payload: Mapping[str, Any],
        request_options: Mapping[str, Any],
        *,
        method: str = "POST",
    ) -> Dict[str, Any]:
        request_id = str(request_options.get("request_id") or uuid.uuid4())
        timeout_s = float(request_options.get("timeout_s", self.timeout_s))
        timeout_s = max(0.001, timeout_s)
        retry = _as_retry_policy(request_options.get("retry"), self.retry)

        headers: MutableMapping[str, str] = {
            "content-type": "application/json",
            "x-request-id": request_id,
        }
        headers.update(self.default_headers)
        headers.update(dict(request_options.get("headers") or {}))

        api_key = request_options.get("api_key", self.api_key)
        if api_key and not _has_header(headers, "x-api-key"):
            headers["x-api-key"] = str(api_key)

        auth_bearer = request_options.get("auth_bearer", self.auth_bearer)
        if auth_bearer and not _has_header(headers, "authorization"):
            token = str(auth_bearer)
            headers["authorization"] = token if token.lower().startswith("bearer ") else f"Bearer {token}"

        admin_token = request_options.get("admin_token", self.admin_token)
        if admin_token and not _has_header(headers, "x-admin-token"):
            headers["x-admin-token"] = str(admin_token)

        req_method = method.upper()
        url = _join_url(self.base_url, path)
        body: Optional[bytes] = None
        if req_method == "GET":
            qp: Dict[str, Any] = {}
            for k, v in payload.items():
                if v is None:
                    continue
                if isinstance(v, bool):
                    qp[k] = "true" if v else "false"
                else:
                    qp[k] = str(v)
            if qp:
                url = f"{url}?{urlencode(qp)}"
        else:
            body = json.dumps(payload).encode("utf-8")

        last_exc: Optional[BaseException] = None

        for attempt in range(retry.max_retries + 1):
            req = request.Request(url=url, data=body, headers=dict(headers), method=req_method)
            try:
                with request.urlopen(req, timeout=timeout_s) as resp:
                    raw = resp.read()
                    status = int(resp.getcode())
                    content_type = resp.headers.get("content-type", "")
                    parsed = _parse_body(raw, content_type)
                    response_request_id = resp.headers.get("x-request-id") or request_id

                    if 200 <= status < 300:
                        return {
                            "data": parsed,
                            "status": status,
                            "request_id": response_request_id,
                        }

                    code = f"http_{status}"
                    message = f"request failed with status {status}"
                    details = None
                    issues = None
                    if isinstance(parsed, dict):
                        code = str(parsed.get("error") or code)
                        message = str(parsed.get("message") or message)
                        details = parsed.get("details")
                        maybe_issues = parsed.get("issues")
                        if isinstance(maybe_issues, list):
                            issues = maybe_issues

                    api_err = AionisApiError(status, code, message, details, response_request_id, issues)
                    can_retry = attempt < retry.max_retries and _should_retry_status(status)
                    if can_retry:
                        retry_after = _parse_retry_after_seconds(resp.headers)
                        time.sleep(retry_after if retry_after is not None else _compute_backoff_seconds(retry, attempt + 1))
                        continue
                    raise api_err

            except error.HTTPError as e:
                raw = e.read() if hasattr(e, "read") else b""
                status = int(e.code)
                content_type = e.headers.get("content-type", "") if e.headers else ""
                parsed = _parse_body(raw, content_type)
                response_request_id = (e.headers.get("x-request-id") if e.headers else None) or request_id

                code = f"http_{status}"
                message = f"request failed with status {status}"
                details = None
                issues = None
                if isinstance(parsed, dict):
                    code = str(parsed.get("error") or code)
                    message = str(parsed.get("message") or message)
                    details = parsed.get("details")
                    maybe_issues = parsed.get("issues")
                    if isinstance(maybe_issues, list):
                        issues = maybe_issues

                api_err = AionisApiError(status, code, message, details, response_request_id, issues)
                can_retry = attempt < retry.max_retries and _should_retry_status(status)
                if can_retry:
                    retry_after = _parse_retry_after_seconds(e.headers or {})
                    time.sleep(retry_after if retry_after is not None else _compute_backoff_seconds(retry, attempt + 1))
                    continue
                raise api_err
            except (error.URLError, TimeoutError, OSError) as e:
                last_exc = e
                if attempt < retry.max_retries:
                    time.sleep(_compute_backoff_seconds(retry, attempt + 1))
                    continue
                break

        raise AionisNetworkError(
            f"network request failed for {req_method} {path}: {str(last_exc) if last_exc else 'unknown error'}",
            request_id=request_id,
        )



from .client import (
    AionisApiError,
    AionisClient,
    AionisNetworkError,
    is_backend_capability_unsupported_error,
    is_shadow_dual_write_strict_failure_error,
    parse_backend_capability_error_details,
    parse_shadow_dual_write_strict_failure_details,
)
from .types import (
    AionisResponse,
    BackendCapabilityErrorDetails,
    CapabilityContractSpec,
    CapabilityFailureMode,
    ControlAlertDeliveriesQuery,
    ControlAlertRouteInput,
    ControlAlertRoutesQuery,
    ControlAlertRouteStatusInput,
    ControlApiKeyInput,
    ControlApiKeysQuery,
    ControlApiKeysStaleQuery,
    ControlApiKeyRotateInput,
    ControlAuditEventsQuery,
    ControlIncidentPublishJobInput,
    ControlIncidentPublishJobsQuery,
    ControlIncidentPublishReplayInput,
    ControlIncidentPublishRollupQuery,
    ControlIncidentPublishSloQuery,
    ControlProjectInput,
    ControlTenantDiagnosticsQuery,
    ControlTenantInput,
    ControlTenantKeyUsageQuery,
    ControlTenantQuotaInput,
    ControlTenantsQuery,
    ControlTenantTimeseriesQuery,
    DecisionLinkMode,
    FeedbackOutcome,
    HealthResponse,
    MemoryEdgeInput,
    MemoryEventWriteInput,
    MemoryFindInput,
    MemoryLane,
    MemoryNodeInput,
    MemoryPackExportInput,
    MemoryPackImportInput,
    MemoryPackV1,
    MemoryRecallInput,
    MemoryRecallTextInput,
    ShadowDualWriteStrictFailureDetails,
    MemorySessionCreateInput,
    MemorySessionEventsListInput,
    MemoryWriteInput,
    RulesEvaluateInput,
    Tier,
    ToolsDecisionInput,
    ToolsDecisionPayload,
    ToolsDecisionResponse,
    ToolsFeedbackInput,
    ToolsFeedbackResponse,
    ToolsSelectDecision,
    ToolsSelectInput,
    ToolsSelectResponse,
)

__all__ = [
    "AionisApiError",
    "AionisClient",
    "AionisNetworkError",
    "AionisResponse",
    "BackendCapabilityErrorDetails",
    "CapabilityContractSpec",
    "CapabilityFailureMode",
    "ControlAlertDeliveriesQuery",
    "ControlAlertRouteInput",
    "ControlAlertRoutesQuery",
    "ControlAlertRouteStatusInput",
    "ControlApiKeyInput",
    "ControlApiKeysQuery",
    "ControlApiKeysStaleQuery",
    "ControlApiKeyRotateInput",
    "ControlAuditEventsQuery",
    "ControlIncidentPublishJobInput",
    "ControlIncidentPublishJobsQuery",
    "ControlIncidentPublishReplayInput",
    "ControlIncidentPublishRollupQuery",
    "ControlIncidentPublishSloQuery",
    "ControlProjectInput",
    "ControlTenantDiagnosticsQuery",
    "ControlTenantInput",
    "ControlTenantKeyUsageQuery",
    "ControlTenantQuotaInput",
    "ControlTenantsQuery",
    "ControlTenantTimeseriesQuery",
    "DecisionLinkMode",
    "FeedbackOutcome",
    "HealthResponse",
    "MemoryEdgeInput",
    "MemoryEventWriteInput",
    "MemoryFindInput",
    "MemoryLane",
    "MemoryNodeInput",
    "MemoryPackExportInput",
    "MemoryPackImportInput",
    "MemoryPackV1",
    "MemoryRecallInput",
    "MemoryRecallTextInput",
    "ShadowDualWriteStrictFailureDetails",
    "MemorySessionCreateInput",
    "MemorySessionEventsListInput",
    "MemoryWriteInput",
    "RulesEvaluateInput",
    "Tier",
    "ToolsDecisionInput",
    "ToolsDecisionPayload",
    "ToolsDecisionResponse",
    "ToolsFeedbackInput",
    "ToolsFeedbackResponse",
    "ToolsSelectDecision",
    "ToolsSelectInput",
    "ToolsSelectResponse",
    "is_backend_capability_unsupported_error",
    "is_shadow_dual_write_strict_failure_error",
    "parse_backend_capability_error_details",
    "parse_shadow_dual_write_strict_failure_details",
]



from __future__ import annotations

from typing import Any, Dict, List, Literal, Optional, TypedDict


MemoryLane = Literal["private", "shared"]
Tier = Literal["hot", "warm", "cold", "archive"]
FeedbackOutcome = Literal["positive", "negative", "neutral"]
DecisionLinkMode = Literal["provided", "inferred", "created_from_feedback"]
CapabilityFailureMode = Literal["hard_fail", "soft_degrade"]


class AionisResponse(TypedDict, total=False):
    data: Any
    status: int
    request_id: Optional[str]


class CapabilityContractSpec(TypedDict, total=False):
    failure_mode: CapabilityFailureMode
    degraded_modes: List[str]


class BackendCapabilityErrorDetails(TypedDict, total=False):
    capability: str
    backend: str
    failure_mode: CapabilityFailureMode
    degraded_mode: str
    fallback_applied: bool


class ShadowDualWriteStrictFailureDetails(TypedDict, total=False):
    capability: Literal["shadow_mirror_v2"]
    failure_mode: CapabilityFailureMode
    degraded_mode: Literal["capability_unsupported", "mirror_failed"]
    fallback_applied: bool
    strict: bool
    mirrored: bool
    error: str


class HealthResponse(TypedDict, total=False):
    ok: bool
    memory_store_backend: str
    memory_store_recall_capabilities: Dict[str, bool]
    memory_store_write_capabilities: Dict[str, bool]
    memory_store_feature_capabilities: Dict[str, bool]
    memory_store_capability_contract: Dict[str, CapabilityContractSpec]


class EdgeEndpointInput(TypedDict, total=False):
    id: str
    client_id: str


class MemoryNodeInput(TypedDict, total=False):
    id: str
    client_id: str
    scope: str
    type: str
    tier: Tier
    memory_lane: MemoryLane
    producer_agent_id: str
    owner_agent_id: str
    owner_team_id: str
    title: str
    text_summary: str
    slots: Dict[str, Any]
    raw_ref: str
    evidence_ref: str
    embedding: List[float]
    embedding_model: str
    salience: float
    importance: float
    confidence: float


class MemoryEdgeInput(TypedDict, total=False):
    id: str
    scope: str
    type: str
    src: EdgeEndpointInput
    dst: EdgeEndpointInput
    weight: float
    confidence: float
    decay_rate: float


class MemoryWriteInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    actor: str
    parent_commit_id: str
    input_text: str
    input_sha256: str
    model_version: str
    prompt_version: str
    auto_embed: bool
    force_reembed: bool
    memory_lane: MemoryLane
    producer_agent_id: str
    owner_agent_id: str
    owner_team_id: str
    trigger_topic_cluster: bool
    topic_cluster_async: bool
    nodes: List[MemoryNodeInput]
    edges: List[MemoryEdgeInput]


class MemoryRecallInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    query_embedding: List[float]
    consumer_agent_id: str
    consumer_team_id: str
    limit: int
    neighborhood_hops: int
    return_debug: bool
    include_embeddings: bool
    include_meta: bool
    include_slots: bool
    include_slots_preview: bool
    slots_preview_keys: int
    max_nodes: int
    max_edges: int
    ranked_limit: int
    min_edge_weight: float
    min_edge_confidence: float
    context_token_budget: int
    context_char_budget: int
    context_compaction_profile: Literal["balanced", "aggressive"]
    rules_context: Dict[str, Any]
    rules_include_shadow: bool
    rules_limit: int


class MemoryRecallTextInput(MemoryRecallInput, total=False):
    query_text: str


class MemoryFindInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    uri: str
    id: str
    client_id: str
    type: Literal["event", "entity", "topic", "rule", "evidence", "concept", "procedure", "self_model"]
    title_contains: str
    text_contains: str
    memory_lane: MemoryLane
    slots_contains: Dict[str, Any]
    consumer_agent_id: str
    consumer_team_id: str
    include_meta: bool
    include_slots: bool
    include_slots_preview: bool
    slots_preview_keys: int
    limit: int
    offset: int


class MemorySessionCreateInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    actor: str
    session_id: str
    title: str
    text_summary: str
    input_text: str
    metadata: Dict[str, Any]
    auto_embed: bool
    memory_lane: MemoryLane
    producer_agent_id: str
    owner_agent_id: str
    owner_team_id: str


class MemoryEventWriteInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    actor: str
    session_id: str
    event_id: str
    title: str
    text_summary: str
    input_text: str
    metadata: Dict[str, Any]
    auto_embed: bool
    memory_lane: MemoryLane
    producer_agent_id: str
    owner_agent_id: str
    owner_team_id: str
    edge_weight: float
    edge_confidence: float


class MemorySessionEventsListInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    consumer_agent_id: str
    consumer_team_id: str
    include_meta: bool
    include_slots: bool
    include_slots_preview: bool
    slots_preview_keys: int
    limit: int
    offset: int


class MemoryPackExportInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    include_nodes: bool
    include_edges: bool
    include_commits: bool
    include_meta: bool
    max_rows: int


class MemoryPackV1(TypedDict, total=False):
    version: Literal["aionis_pack_v1"]
    tenant_id: str
    scope: str
    nodes: List[Dict[str, Any]]
    edges: List[Dict[str, Any]]
    commits: List[Dict[str, Any]]


class MemoryPackImportInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    actor: str
    verify_only: bool
    auto_embed: bool
    manifest_sha256: str
    pack: MemoryPackV1


class RulesEvaluateInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    context: Dict[str, Any]
    include_shadow: bool
    limit: int


class ToolsSelectInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    run_id: str
    context: Dict[str, Any]
    candidates: List[str]
    include_shadow: bool
    rules_limit: int
    strict: bool


class ToolsDecisionInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    decision_id: str


class ToolsFeedbackInput(TypedDict, total=False):
    tenant_id: str
    scope: str
    actor: str
    run_id: str
    decision_id: str
    outcome: FeedbackOutcome
    context: Dict[str, Any]
    candidates: List[str]
    selected_tool: str
    include_shadow: bool
    rules_limit: int
    target: Literal["tool", "all"]
    note: str
    input_text: str
    input_sha256: str


class ToolsSelectDeniedItem(TypedDict, total=False):
    name: str
    reason: str


class ToolsSelectDecision(TypedDict, total=False):
    decision_id: str
    run_id: Optional[str]
    selected_tool: Optional[str]
    policy_sha256: str
    source_rule_ids: List[str]
    created_at: Optional[str]


class ToolsSelectSelection(TypedDict, total=False):
    candidates: List[str]
    selected: Optional[str]
    ordered: List[str]
    denied: List[ToolsSelectDeniedItem]
    fallback: Dict[str, Any]


class ToolsSelectResponse(TypedDict, total=False):
    tenant_id: str
    scope: str
    candidates: List[str]
    selection: ToolsSelectSelection
    rules: Dict[str, Any]
    decision: ToolsSelectDecision


class ToolsFeedbackResponse(TypedDict, total=False):
    ok: bool
    tenant_id: str
    scope: str
    updated_rules: int
    rule_node_ids: List[str]
    commit_id: Optional[str]
    commit_hash: Optional[str]
    decision_id: str
    decision_link_mode: DecisionLinkMode
    decision_policy_sha256: str


class ToolsDecisionPayload(TypedDict, total=False):
    decision_id: str
    decision_kind: Literal["tools_select"]
    run_id: Optional[str]
    selected_tool: Optional[str]
    candidates: List[str]
    context_sha256: str
    policy_sha256: str
    source_rule_ids: List[str]
    metadata: Dict[str, Any]
    created_at: str
    commit_id: Optional[str]


class ToolsDecisionResponse(TypedDict, total=False):
    tenant_id: str
    scope: str
    decision: ToolsDecisionPayload


class ControlTenantInput(TypedDict, total=False):
    tenant_id: str
    display_name: Optional[str]
    status: Literal["active", "suspended"]
    metadata: Dict[str, Any]


class ControlTenantsQuery(TypedDict, total=False):
    status: Literal["active", "suspended"]
    limit: int
    offset: int


class ControlProjectInput(TypedDict, total=False):
    project_id: str
    tenant_id: str
    display_name: Optional[str]
    status: Literal["active", "archived"]
    metadata: Dict[str, Any]


class ControlApiKeyInput(TypedDict, total=False):
    tenant_id: str
    project_id: Optional[str]
    label: Optional[str]
    role: Optional[str]
    agent_id: Optional[str]
    team_id: Optional[str]
    metadata: Dict[str, Any]


class ControlApiKeysQuery(TypedDict, total=False):
    tenant_id: str
    project_id: str
    status: Literal["active", "revoked"]
    limit: int
    offset: int


class ControlApiKeysStaleQuery(TypedDict, total=False):
    max_age_days: int
    warn_age_days: int
    rotation_window_days: int
    limit: int


class ControlApiKeyRotateInput(TypedDict, total=False):
    label: Optional[str]
    metadata: Dict[str, Any]


class ControlAlertRouteInput(TypedDict, total=False):
    tenant_id: str
    channel: Literal["webhook", "slack_webhook", "pagerduty_events"]
    label: Optional[str]
    events: List[str]
    status: Literal["active", "disabled"]
    target: str
    secret: Optional[str]
    headers: Dict[str, str]
    metadata: Dict[str, Any]


class ControlAlertRoutesQuery(TypedDict, total=False):
    tenant_id: str
    channel: Literal["webhook", "slack_webhook", "pagerduty_events"]
    status: Literal["active", "disabled"]
    limit: int
    offset: int


class ControlAlertRouteStatusInput(TypedDict, total=False):
    status: Literal["active", "disabled"]


class ControlAlertDeliveriesQuery(TypedDict, total=False):
    tenant_id: str
    event_type: str
    status: Literal["sent", "failed", "skipped"]
    limit: int
    offset: int


class ControlIncidentPublishJobInput(TypedDict, total=False):
    tenant_id: str
    run_id: str
    source_dir: str
    target: str
    max_attempts: int
    metadata: Dict[str, Any]


class ControlIncidentPublishJobsQuery(TypedDict, total=False):
    tenant_id: str
    status: Literal["pending", "processing", "succeeded", "failed", "dead_letter"]
    limit: int
    offset: int


class ControlIncidentPublishReplayInput(TypedDict, total=False):
    tenant_id: str
    statuses: List[Literal["failed", "dead_letter"]]
    ids: List[str]
    limit: int
    reset_attempts: bool
    reason: str
    dry_run: bool
    allow_all_tenants: bool


class ControlTenantQuotaInput(TypedDict, total=False):
    recall_rps: float
    recall_burst: int
    write_rps: float
    write_burst: int
    write_max_wait_ms: int
    debug_embed_rps: float
    debug_embed_burst: int
    recall_text_embed_rps: float
    recall_text_embed_burst: int
    recall_text_embed_max_wait_ms: int


class ControlAuditEventsQuery(TypedDict, total=False):
    tenant_id: str
    action: str
    limit: int
    offset: int


class ControlTenantDiagnosticsQuery(TypedDict, total=False):
    scope: str
    window_minutes: int


class ControlIncidentPublishRollupQuery(TypedDict, total=False):
    window_hours: int
    sample_limit: int


class ControlIncidentPublishSloQuery(TypedDict, total=False):
    window_hours: int
    baseline_hours: int
    min_jobs: int
    adaptive_multiplier: float
    failure_rate_floor: float
    dead_letter_rate_floor: float
    backlog_warning_abs: int
    dead_letter_backlog_warning_abs: int
    dead_letter_backlog_critical_abs: int


class ControlTenantTimeseriesQuery(TypedDict, total=False):
    endpoint: Literal["write", "recall", "recall_text"]
    window_hours: int
    limit: int
    offset: int
    cursor: str


class ControlTenantKeyUsageQuery(TypedDict, total=False):
    endpoint: Literal["write", "recall", "recall_text"]
    window_hours: int
    baseline_hours: int
    min_requests: int
    zscore_threshold: float
    limit: int
    offset: int
    cursor: str


__all__ = [
    "AionisResponse",
    "ControlAlertDeliveriesQuery",
    "ControlAlertRouteInput",
    "ControlAlertRoutesQuery",
    "ControlAlertRouteStatusInput",
    "ControlApiKeyInput",
    "ControlApiKeysQuery",
    "ControlApiKeysStaleQuery",
    "ControlApiKeyRotateInput",
    "ControlAuditEventsQuery",
    "ControlIncidentPublishJobInput",
    "ControlIncidentPublishJobsQuery",
    "ControlIncidentPublishReplayInput",
    "ControlIncidentPublishRollupQuery",
    "ControlIncidentPublishSloQuery",
    "ControlProjectInput",
    "ControlTenantDiagnosticsQuery",
    "ControlTenantInput",
    "ControlTenantKeyUsageQuery",
    "ControlTenantQuotaInput",
    "ControlTenantsQuery",
    "ControlTenantTimeseriesQuery",
    "DecisionLinkMode",
    "FeedbackOutcome",
    "MemoryEdgeInput",
    "MemoryEventWriteInput",
    "MemoryFindInput",
    "MemoryLane",
    "MemoryNodeInput",
    "MemoryPackExportInput",
    "MemoryPackImportInput",
    "MemoryPackV1",
    "MemoryRecallInput",
    "MemoryRecallTextInput",
    "MemorySessionCreateInput",
    "MemorySessionEventsListInput",
    "MemoryWriteInput",
    "RulesEvaluateInput",
    "Tier",
    "ToolsDecisionInput",
    "ToolsDecisionPayload",
    "ToolsDecisionResponse",
    "ToolsFeedbackInput",
    "ToolsFeedbackResponse",
    "ToolsSelectDecision",
    "ToolsSelectInput",
    "ToolsSelectResponse",
]
